import .fintype
import .stlc_num
import .tactics

/- Full reduction. Star, typing as in WN. -/
open nat tm

reserve infixl `≻`:40
inductive eval : Π {n}, tm n → tm n → Prop
infixl `≻` := eval
| eappl {n} {e₁ : tm n} {e₁' e₂} : e₁ ≻ e₁' →  app e₁ e₂ ≻ app e₁' e₂
| eappr {n} {e₁ : tm n} {e₂ e₂'} : e₂ ≻ e₂' → app e₁ e₂ ≻ app e₁ e₂'
| elam {n} {e1 : tm (nat.succ n)} {e2} : e1 ≻ e2 → lam e1 ≻ lam e2
| eplusl {n} {e₁ : tm n}{e₁' e₂} : e₁ ≻ e₁' → plus e₁ e₂ ≻ plus e₁' e₂
| eplusr {n} {e₁ : tm n} {e₂ e₂'} : e₂ ≻ e₂' → plus e₁ e₂ ≻ plus e₁ e₂'
| econst {n} {n₁ n₂} : @eval n (plus (const n₁) (const n₂)) (const (n₁+n₂))
| ebeta {n} {e₁ : tm (nat.succ n)} {e₂} : app (lam e₁) e₂ ≻ (subst_tm (scons e₂ var_tm ) (e₁))
infix `≻` := eval
open eval

meta def HintsEval:= [``(eappl), ``(eappr), ``(elam), ``(eplusl), ``(eplusr), ``(econst), ``(ebeta)]

inductive star {n} (R : tm n → tm n → Prop) : tm n → (tm n → Prop)
| srefl {e} : star e e 
| sstep {e₁ e₂ e₃} :  (R e₁ e₂) → star e₂ e₃ → star e₁ e₃
open star
infix `≻*`:40 := star eval

inductive type : Type
| tint : type
| tarrow : type → type → type
open type
infixl `⤏`:40 := tarrow

inductive types : Π {m}, (Fin m → type) → tm m → type → Prop
| tvar {m} Γ (x : Fin m) : types Γ (var_tm x) (Γ x)
| tapp {m} Γ (e₁ : tm m) e₂ (A B) : types Γ e₁ (A ⤏ B) → types Γ e₂ A → types Γ (app e₁ e₂) B
| tplus {m} Γ (e₁ : tm m) (e₂) : types Γ e₁ tint → types Γ e₂ tint → types Γ (plus e₁ e₂) tint 
| tlam {m} Γ (e : tm (nat.succ m)) (A B) : types (@scons _ m  A Γ) e B → types Γ (lam e) (A ⤏ B)
| tint {m} (Γ n) : @types m Γ (const n) tint 
notation Γ `⊢`:50 x `:` A:50 := types Γ x A. 
open types


definition ctx (m) := Fin m → type
lemma ev_to_star {n} (e1 : tm n) (e2) : e1 ≻ e2 → e1 ≻* e2  :=
by intros; apply sstep; aauto_w ``(srefl)


lemma star_trans {n} (e1 : tm n) (e2 e3) : e1 ≻* e2 → e2 ≻* e3 → e1 ≻* e3  :=
begin intros h₁ h₂,induction h₁, 
    { aauto }, 
    { apply sstep; aauto },
end

meta def HintsStar := [``(star.srefl), ``(star.sstep), ``(star_trans), ``(ev_to_star)]

def empty_ctx : Fin 0 → type :=  λ (x : Fin 0), match x with end 
notation `∅`:= empty_ctx

/- Agreement of contexts. -/
definition agree_ren {n m} (Γ : ctx n) (Γ' : ctx m) (ξ : Fin m → Fin n) 
    := forall x, (Γ (ξ x)) = Γ' x. 

definition agree_subst_types {n m} (Γ : ctx n) (Γ' : ctx m) (σ : Fin n → tm m) 
    := forall x, types Γ' (σ x) (Γ x). -- another version with R will be defined later (agree_subst)  


notation Γ `≼`:50 Δ `:`:50 ξ := agree_ren Δ Γ ξ.

/- Morphism lemmas. -/
lemma morphism_ren {m} (Γ : ctx m) (x : tm m) (A) (H : Γ ⊢ x : A ) :  
  ∀ m' Δ (ξ : Fin m → Fin m'), (Γ ≼ Δ : ξ) → 
   Δ ⊢ (ren_tm  ξ x) : A :=
begin  induction H ; intros; aauto, 
  { arw, specialize (a H_x), rw ← a, constructor },
  { arw, constructor; aauto },
  { arw, constructor; aauto }, 
  { arw1 , arw1, constructor,
    apply H_ih, intro x, 
    cases x, { reflexivity}, 
    { specialize(a x_a), generalize h_3 : H_Γ x_a = a_3, generalize h_4 : Δ (ξ x_a) = a_4,
    revert a , rw h_3, rw h_4, intro a, cases a, 
    simp [scons], 
    rw h_3, rw ← h_4,  reflexivity },},
    { arw, constructor; aauto }
end


lemma morphism_subst {m} (Γ : ctx m) (x : tm m) (A) (H : types Γ x A) :
  ∀ m' Δ  (σ : Fin m → tm m'), (agree_subst_types Γ Δ σ) 
  → Δ  ⊢ (subst_tm σ x) : A := 
begin induction H; intros;  any_goals {arw, aauto, try { constructor; aauto},}, 
  intro x, cases x, 
    { constructor}, 
    { apply morphism_ren, aauto, intro x, aauto } 
 end


set_option trace.check true

/- Substitutivity. -/
lemma substitutivity_h {n} (e1 : tm (succ n)) (e2 t) :
    t = subst_tm (scons e2 var_tm) e1 → (app (lam e1) e2) ≻ t :=
begin
    intro h, rw h, constructor, 
end
  
lemma substitutivity {n m} (s t σ) : s ≻ t → s[σ] ≻ @subst_tm n m σ t :=
begin 
    intro h, revert m σ , induction h; intros m σ, 
    any_goals { arw, constructor; aauto },
    { arw,   
     apply substitutivity_h, arw, simp [scons],
      }
end



lemma substitutivity_ren {n m} (s t ξ) : s ≻ t → s.⟨ξ⟩ ≻ @ren_tm n m ξ t :=
begin 
    intro h, revert m ξ, induction h; intros m σ, 
    any_goals { arw, constructor; aauto },
    { arw, apply substitutivity_h, arw, 
    simp[var_zero, scons], },
end



lemma star_subst {n m} (s t) (σ : Fin n -> tm m) : s ≻* t → s[σ] ≻* t[σ] :=
begin 
    intro h, revert σ, induction h; intros σ, 
    { apply srefl },
    { apply sstep, apply substitutivity, all_goals{ aauto },  }
end

lemma star_ren {n m} (s t) (ξ : Fin n -> Fin m) : s ≻* t -> s.⟨ξ⟩ ≻* t.⟨ξ⟩ :=

begin intros h , revert ξ, induction h; intros ξ,
    { apply srefl, },
    { apply sstep, apply substitutivity_ren, all_goals {aauto}}
end

lemma eval_appl {n} (e1 : tm n) (e1' e2) : e1 ≻* e1' → (app e1 e2) ≻* (app e1' e2) :=
begin 
    intro h, induction h, constructor, apply sstep, apply eappl, aauto, aauto, 
end

lemma eval_appr {n} (e1 : tm n) (e2' e2) : e2 ≻* e2' → (app e1 e2) ≻* (app e1 e2') :=
begin 
    intro h, induction h, constructor, apply sstep, apply eappr, aauto, aauto, 
end

lemma eval_app {n} (e1 : tm n) (e1' e2 e2') : e1 ≻* e1' → e2 ≻* e2' → (app e1 e2) ≻* (app e1' e2'):=
begin
    intros h₁ h₂ , apply star_trans, apply eval_appl; aauto, apply eval_appr, aauto,
end
lemma eval_plusl {n} (e1 : tm n) (e1' e2) : e1 ≻* e1' → (plus e1 e2) ≻* (plus e1' e2) :=
begin 
    intro h, induction h, constructor, apply sstep, apply eplusl, aauto, aauto, 
end

lemma eval_plusr {n} (e1 : tm n) (e2' e2) : e2 ≻* e2' → (plus e1 e2) ≻* (plus e1 e2') :=
begin 
    intro h, induction h, constructor, apply sstep, apply eplusr, aauto, aauto, 
end

lemma eval_plus {n} (e1 : tm n) (e1' e2 e2') : e1 ≻* e1' → e2 ≻* e2' → (plus e1 e2) ≻* (plus e1' e2'):=
begin
    intros h₁ h₂ , apply star_trans, apply eval_plusl; aauto, apply eval_plusr, aauto,
end

lemma eval_lam {n} (e1 : tm (succ n)) (e2) : e1 ≻* e2 → lam e1 ≻* lam e2 :=
begin
    intro h, induction h, apply srefl, constructor, apply elam, all_goals{aauto},
end 

lemma subst_eval {n m} (s) (σ τ : Fin n → tm m) : (forall x, σ x ≻* τ x) → s[σ] ≻* s[τ] :=
begin 
    revert m σ τ, induction s; intros,
    { arw, aauto },
    { arw, apply eval_app; aauto,},
    { arw, apply eval_lam, aauto, intro x, cases x,
        { apply srefl },
        { simp [scons], specialize (a x_a), 
        revert a, generalize h3 : (σ x_a)= a_3, generalize h4 : (τ x_a)=a_4,
        intro a_1, cases a_1, subst h3, apply star_ren, rw h4, apply srefl,
        apply star_ren, apply sstep, rw h3, aauto, rw h4, aauto
        }
    },
    { arw, apply srefl },
    { arw, apply eval_plus; aauto  }
end

lemma subst_eval_head {n} (s : tm (succ n)) (t t') : t ≻ t' → s[t..] ≻* s[t'..] := 
begin 
    intro ht, apply subst_eval, intro x, cases x; simp [scons], 
    all_goals { aautow HintsStar }, 
    aautow HintsStar
end

lemma preservation {n} (Γ) (s : tm n) (t) : s ≻ t → forall A, Γ ⊢ s : A → Γ ⊢ t : A :=
begin 
    intro hs, induction hs; intros A ht; cases ht; try { constructor}; aauto, 
    { cases ht_a, apply morphism_subst, aauto, 
    intro x, cases x, aauto, simp[scons], constructor }
end     

lemma agree_weaken {n m} (Γ : ctx n) (Δ : ctx m) (A : type) (ξ) :
( Γ ≼ Δ : ξ )→ (Γ ≼ scons A Δ : comp shift ξ) := 
begin 
    intros h x, rw ← (h x), simp [scons], reflexivity, 
end


lemma types_ren {n m} (Γ) (s : tm n) (ξ : Fin n → Fin m) (Δ : ctx m) (A) : 
    Γ ⊢ s : A → (Γ ≼ Δ : ξ) → Δ ⊢ s.⟨ξ⟩ : A :=
begin 
    intro h, revert Δ ξ m, induction h; intros m' Δ ξ p;
    try {arw, constructor; aauto }, --simp with substLemmas; constructor; aauto },
    { rw ← (p h_x), arw, constructor  },
    { intro x, cases x, refl, aauto  }
end
