import .fintype
import .stlc_num
import .tactics

/- Call-by-value reduction, star. -/

open tm
definition value {n} : tm n → Prop :=
  λ e, match e with 
  | const n := true
  | lam e := true
  | _ := false
  end

notation `Λ`:50 := lam 
reserve infixl `≻`:50
inductive step {n}: tm n → tm n → Prop 
infixl ` ≻ `:= step
  | appl : forall {e₁} {e₁'} {e₂} , e₁ ≻ e₁' → app e₁ e₂ ≻ app e₁' e₂
  | appr : forall {e₁} {e₂} {e₂'}, value e₁ → e₂ ≻ e₂' → app e₁ e₂ ≻ app e₁ e₂'
  | plusl : forall {e₁} {e₁'} {e₂} , e₁ ≻ e₁' → plus e₁ e₂ ≻ plus e₁' e₂
  | plusr : forall {e₁} {e₂} {e₂'}, value e₁ → e₂ ≻ e₂' → plus e₁ e₂ ≻ plus e₁ e₂'
  | splus : forall {n₁} {n₂}, plus (const n₁) (const n₂) ≻ const (n₁ + n₂)
  | sbeta : forall {e₁} {e₂}, value e₂ → app (lam e₁) e₂ ≻ (subst_tm (scons e₂ var_tm) e₁)
infixl ` ≻ `:= step
open step

inductive star {n} (R : tm n → tm n → Prop) : tm n → (tm n → Prop)
  | srefl {e} : star e e 
  | sstep {e₁ e₂ e₃} :  (R e₁ e₂) → star e₂ e₃ → star e₁ e₃
open star

infixl ` ≻* `:50 := star step

lemma star_trans {n} (e₁ : tm n) (e₂ e₃) : e₁ ≻* e₂ → e₂ ≻* e₃ → e₁ ≻* e₃ :=
begin
  intros h₁ h₂, induction h₁; aauto; aauto_w ``(sstep) 
end  

lemma to_star {n} (e₁ e₂ : tm n) : e₁ ≻ e₂ → e₁ ≻* e₂ := by intro H; apply sstep; try{ assumption }; apply srefl 

/- Typing. -/

inductive type : Type
  | tint : type
  | tarrow : type → type → type
open type
infixl `⤏`:50 := tarrow


definition ctx (m) := Fin m → type

inductive types : Π {m}, (Fin m → type) → tm m → type → Prop
  | tvar {m} Γ (x : Fin m) : types Γ (var_tm x) (Γ x)
  | tapp {m} Γ (e₁ : tm m) e₂ (A B) : types Γ e₁ (tarrow A B) → types Γ e₂ A → types Γ (app e₁ e₂) B
  | tplus {m} Γ (e₁ : tm m) (e₂) : types Γ e₁ tint → types Γ e₂ tint → types Γ (plus e₁ e₂) tint 
  | tlam {m} Γ (e : tm (nat.succ m)) (A B) : types (@scons _ m  A Γ) e B → types Γ (lam e) (A ⤏ B)
  | tconst {m n} (Γ : ctx m) : types Γ (const n) tint
notation Γ ` ⊢ `:50 x ` : ` A:50 := types Γ x A. 
open types

def empty_ctx : Fin 0 → type :=  λ (x : Fin 0), match x with end 
notation `∅`:= empty_ctx

/- Morphism, agreement and preservation properties. (not all needed for WN) -/
definition agree_ren {n m} (Δ : ctx m) (Γ : ctx n) (ξ : Fin m → Fin n) :=
  ∀ x, Γ (ξ x) = Δ x.
notation Δ ` ≺ `:50 Γ ` ; ` ξ:50 := agree_ren Δ Γ ξ.

definition agree_subst {n m} (Δ : ctx m) (Γ : ctx n) (σ : Fin m → tm n) :=
  ∀ x, Γ ⊢ (σ x) : Δ x .
notation Δ ` ≺ `:50 Γ ` : `σ:50 := agree_subst Δ Γ σ. 

lemma morphism_ren {n} (Γ : ctx n) (s) (A) :
  Γ ⊢ s : A → ∀ {m} Δ (ζ : Fin n → Fin m), 
    (Γ ≺ Δ ; ζ) → Δ ⊢ (ren_tm ζ s) : A := 
begin 
  intros H₁, induction H₁; intros; simp with substLemmas at *;
  try {constructor}; aauto,  
  { rw ← (a H₁_x), constructor },
  intro x, cases x, { refl }, unfold scons ,rw ← (a x_a), refl
end


lemma morphism_subst {n} (Γ : ctx n) (s) (A) :
  Γ ⊢ s : A → ∀ {m} Δ (σ : Fin n → tm m), 
    (Γ ≺ Δ : σ ) → Δ ⊢ (subst_tm σ s) : A := 
begin
  intros H₁, induction H₁; intros; simp with substLemmas; arw, 
  { aauto },
  { constructor; aauto,  },
  { constructor; aauto }, 
  { unfold comp, constructor, apply H₁_ih, 
  intros x, cases x with _ _ x' , {  constructor },
  { specialize (a x'), apply morphism_ren, unfold scons, aauto,
  intro x, aauto } }, constructor
end


-- Note: A generalization step is necessary before an induction is possible,
-- we have to generalize the index 0 in the type of the empty context
lemma preservation (A : type) : ∀ (e₁ : tm 0),
  empty_ctx ⊢ e₁ : A → ∀ e₂, e₁ ≻ e₂ → empty_ctx ⊢ e₂ : A :=
suffices ∀ {n}, n = 0 → ∀ {ctx} (e₁ : tm n), types ctx e₁ A → ∀ e₂, step e₁ e₂ → types ctx e₂ A,
from this rfl,
begin
  introv h H₁ H₂,
  induction H₁ generalizing h; subst h,
  { cases H₂ },
  { cases H₂, any_goals { constructor }, all_goals{  aauto_w ``(morphism_subst)},
    {  cases H₁_a, aauto, }, 
    { intro x, cases x, { unfold scons, aauto  }, { constructor }, },
  }, 
  { cases H₂, constructor; aauto, constructor, aauto, { aauto_cautious 0, }, 
    constructor; aauto
      },
  { cases H₂ }, 
  { cases H₂ }
end

/- alternative definition:
lemma preservation' (A : type) (e₁ : tm 0)
  (H₁ : types empty_ctx e₁ A) (e₂) (H₂ : e₁ ≻ e₂) : empty_ctx ⊢ e₂ : A  := by ... -/


/- Properties of star. (Compatibility) -/
lemma star_appl {n} {e₁ : tm n} {e₁' e₂} : 
  e₁ ≻* e₁' → (app e₁ e₂) ≻* (app e₁' e₂) :=
begin 
  intros h, 
  induction h, 
  any_goals { constructor; aauto; aauto_w ``(appl),  }, 
end

lemma star_appr {n} {e₁ : tm n} {e₂ e₂'} : 
  value e₁ → e₂ ≻* e₂' → app e₁ e₂ ≻* app e₁ e₂' :=
begin 
  intros hv h, induction h, 
  { apply srefl, },
  { apply sstep; aauto_w ``(appr) }
end

lemma star_sbeta {n} {e₁ : tm (nat.succ n)} {e₂ : tm n} : 
  value e₂ → app (lam e₁) e₂ ≻* (subst_tm (scons e₂ var_tm) e₁) :=
begin 
  intro hv, constructor; aauto_w ``(sbeta), aauto_w ``(srefl)
end


