open tactic 

/-- Rewriting and expressions. --/

meta def rw_pexpr (e : pexpr) : tactic unit := 
do 
  e ← tactic.to_expr e, 
  t ← target, 
  (p,h,_) ← tactic.rewrite e t,
  replace_target p h

meta def rw_expr (e : expr) : tactic unit := 
do
  t ← target, 
  (p,h,_) ← rewrite e t,
  replace_target p h


-- tries to rewrite the target with pre-expressions [``(lemma_name),...]
-- tracing can be disabled
meta def rw_exprs (default : tactic unit) (trc := tt): list pexpr → tactic unit
| [] := default
| (e::es) := do (rw_pexpr e >> if trc then trace("✓ " ++ to_string e) else skip) <|> rw_exprs es

meta def rw_exprs_trace_all (default : tactic unit) (trc := ff): list pexpr → tactic unit
| [] := default
| (e::es) := do 
          (rw_pexpr e >> if trc then trace("✓ " ++ to_string e) else skip) 
          <|> do if trc then (trace ("✕ " ++ to_string e)) else skip, rw_exprs_trace_all es



-- application: Autosubst's rewriting steps
meta def rw_lemmas (rw_tac : list pexpr → tactic unit) (RwLemmas : list pexpr):= 
  rw_tac RwLemmas

meta def rw_unfoldables (rwtac : list pexpr → tactic unit) (Unfoldables : list pexpr) := 
  rwtac Unfoldables



/-- Some tactics like eauto--/

-- 'Example from Metaprogramming in Lean': find hypothesis with the same type in the context
meta def find_ty : expr → list expr → tactic expr
| e []        := tactic.failed
| e (h :: hs) :=  do t ← tactic.infer_type h, (tactic.unify e t >> return h) <|> find_ty e hs

-- assumption
meta def assmt (e : expr) : tactic expr :=
do ctx ← tactic.local_context, find_ty e ctx

meta def exact_assmt : tactic unit := 
do {t ← tactic.target, r ← assmt t, tactic.exact r} <|> tactic.fail "no such assumption"

meta def exact_assmt_tr : tactic unit := 
do {t ← tactic.target, r ← assmt t, tactic.exact r >> tactic.trace ("exact "++to_string r)} <|> tactic.fail "no such assumption"


-- try to apply a tactic to a list of expressions
meta def try_apply_fun {α} (f : expr → tactic α) (default : tactic unit ) : list _ → tactic unit
| []  := default 
| (h :: hs) :=(f h >> skip) <|> try_apply_fun hs --(f h >> trace ("applied " ++ to_string h)) <|> try_apply_fun hs

-- same for preexpressions
meta def try_apply_fun_pre {α} (f : expr → tactic α) (default : tactic unit ) : list pexpr → tactic unit
| []  := default 
| (h :: hs) := do h ← to_expr h, 
(f h >> trace ("applied hint: " ++ to_string h)) 
<|> (try_apply_fun_pre hs)

-- apply only if tactic "ensure" succeeds afterwards (backtracking)
meta def try_apply_fun_ensure (f : expr → tactic unit) (default ensure : tactic unit ) : list _ → tactic unit
| []  := default 
| (h :: hs) := (f h>>ensure) <|> try_apply_fun_ensure hs


-- apply a hypothesis with matching type in the context
meta def apply_ctx (fail : bool := tt): tactic unit :=
let default := if fail then failed else skip in
  do es ← tactic.local_context , try_apply_fun apply default es


-- modification
meta def apply_cautious {α} (f : expr → tactic α ) (default : tactic unit)
(solve_tactic : ℕ → tactic unit) (depth : ℕ) : list _ → tactic unit 
| [] := default 
| (h::hs) := (f h >> (all_goals $ solve_tactic $ depth+1)) <|> apply_cautious hs

meta def now (t : tactic unit) : tactic unit := 
  t >> (reflexivity <|> triv <|> done) 


-- uses or applies assumptions to the current subgoal if possible
-- tracing can be enabled
meta def aauto' : bool →  tactic unit
| tt := do exact_assmt_tr <|> 
  focus1 (apply_ctx >> (all_goals (aauto' tt <|> triv))) 
    <|> exact_assmt_tr <|> 
    ((trace "nothing to be done") >> (try $ now skip) )
| ff := do exact_assmt <|> 
  focus1 (apply_ctx >> (all_goals (aauto' ff <|> triv))) 
    <|> exact_assmt <|>  (try $ now skip) 

meta def aauto (tr:=ff)  := try intros; aauto' ff

meta def aautor (tr:=ff)  :=  repeat $ aauto' ff

meta def aauto_trace  : tactic unit := try intros; aauto' tt

meta def aautor_trace  : tactic unit := repeat $ aauto' tt


-- cautious version (i.e. only apply if resulting subgoals can be solved recursively)
-- with a seach depth d
meta def aauto_cautious : ℕ → tactic unit
| d := (tactic.local_context >>= apply_cautious apply failed aauto_cautious d) 
       <|> exact_assmt <|> now skip


/-- Hints --/

-- apply a specific expression, e.g. a lemma ``(lemma_name)
meta def try_apply_e (e : expr) : tactic unit := 
try_apply_fun apply (trace "not applied: " >> trace e) [e] 

meta def apply_report {α} (n : string)(e : α) : expr -> tactic unit := λ e, apply e >> trace n >> trace e <|> trace "skipped: " >>trace e


-- aauto which uses some hints (lemmas, constructors, ...)
meta def aauto_using_h (default : tactic unit) : list pexpr → tactic unit 
| [] := default 
| (p::ps) := do 
          e ← to_expr p, 
          try_apply_fun apply skip  [e],
          try $ aauto_using_h ps

 
-- single hint
meta def aauto_w' (tr) ( ps : list pexpr) :  tactic unit :=
do focus1 (aauto_using_h (assumption <|> skip) ps; all_goals (aauto tr))

meta def aauto_w_tr (p : pexpr) : tactic unit := 
  aauto_w' tt [p] 

meta def aauto_w (p : pexpr) : tactic unit := aauto_w' ff [p]



-- more control over what is applied, using backtracking
meta def aautow' : nat → bool →  list pexpr →  tactic unit 
| n b [] := skip
| n tt ps := 
  trace ("lvl "++to_string n) >>
  aauto' ff >> 
  try_apply_fun_pre apply skip (ps)
  >> (all_goals (aauto' ff <|> triv <|> aautow' (nat.succ n) tt ps))
  <|> exact_assmt_tr <|> 
  (trace ("nth "++to_string n) >> (try $ now skip) )--sth is applied only if can be solved recursively, else backtrack
--without tracing: 
| n ff ps := 
  aauto' ff >>
  try_apply_fun_pre apply skip (ps)
  >> (all_goals (aauto' ff <|> triv <|> aautow' (nat.succ n) tt ps))
  <|> exact_assmt_tr <|>  (try $ now skip) 
  
meta def aautow : list pexpr → tactic unit :=
 λ p,(triv <|> focus1 ( aautow' 0 ff p ))


