{-# LANGUAGE FlexibleContexts      #-}
{-# LANGUAGE FlexibleInstances     #-}
{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE TypeSynonymInstances  #-}
module Autosubst.Coq where

-- Pretty Printing of Syntax to Coq
import           Autosubst.Syntax
import           Text.PrettyPrint.Leijen

-- TODO: More documentation what is what.

-- Coq Primitives
eqSym_ :: Term -> Term
eqSym_ s = idApp "eq_sym"  [s]

eqTrans_ :: Term -> Term -> Term
eqTrans_ s t = idApp "eq_trans"  [s,t]

-- NAT
nat :: Term
nat = TermId "nat"

none_ :: String
none_ = "None"

some_ :: String
some_ = "Some"

up_ren_ :: Term
up_ren_ = TermId "up_ren"

succ_ :: Term -> Term
succ_ n = TermApp (TermId "S") [n]

fin_ :: Term -> Term
fin_ n = TermApp (TermId "fin") [n]

-- TODO: Nice generation of >>
(>>>) :: Term -> Term -> Term
s >>> t = idApp "comp" [TermId "tt", t, s]

eq_refl_ :: Term
eq_refl_ = TermId "eq_refl"

shift_ :: Term
shift_ = TermId "shift"

id_ :: Term
id_ = TermId "id"

-- TODO: Show notation.
cons_ :: Term
cons_ = TermId "scons"

varZero_ :: Term
varZero_ = TermId "var_zero"

matchFin_ :: Term -> (Term -> Term) -> Term -> Term
matchFin_ s f b = TermMatch (MatchItem s Nothing Nothing) Nothing [Equation (PatCtor some_ ["n"]) (f (TermId "n")), Equation (PatQualId none_) b]

ap_ :: Terms -> Term
ap_ = idApp "ap"


-- CoqShow Class
class CoqShow a where
  coqShow :: a -> Doc

instance CoqShow String where
  coqShow = text

-- Generation of tuples accepted by Coq, no parentheses if there is just one component.
cTupled ::  [Doc] -> Doc
cTupled []  = empty
cTupled [x] = x
cTupled xs  = encloseSep lparen rparen (comma <+> empty) xs

coqPreamble :: Doc
coqPreamble = text "Require Export fintype."

-- Coq Primitives.
instance CoqShow CBinder where
  coqShow (BinderName s) = coqShow s
  coqShow (BinderNameType s t) = parens $ hsep (map coqShow s) <+> text ":" <+> coqShow t
  coqShow (BinderImplicitName s) = text "{" <+> coqShow s <+> text "}"
  coqShow (BinderImplicitNameType s t) = text "{" <+> hsep (map coqShow s) <+> text ":" <+> coqShow t <+> text "}"

instance CoqShow Definition where
  coqShow (Definition name bs mb s) = text "Definition" <+> coqShow name <+> hsep (map coqShow bs) <+> (text ":" <+> coqShow mb) <+> text ":="
                                        <$$> indent 2 (coqShow s) <> text "."

instance CoqShow Equation where
  coqShow (Equation pat s) = text "|" <+> coqShow pat <+> text "=>" <+> coqShow s

instance CoqShow Fixpoint where
  coqShow (Fixpoint bs) =  text "Fixpoint"
    <+> hcat (punctuate (line <> text " with ") (map (nest 2 . coqShow) bs)) <> dot

instance CoqShow FixpointBody where
  coqShow (FixpointBody name bs ret s) = coqShow name <+> hsep (map coqShow bs) <+> text ":" <+> coqShow ret
                                            <+> text ":=" <$$> indent 2 (coqShow s)

instance CoqShow Lemma where
  coqShow (Lemma name bs mb s) = text "Lemma" <+> coqShow name <+> hsep (map coqShow bs) <+> (text ":" <+> coqShow mb) <+> text "."
                                        <$$> text "Proof." <+> coqShow s <+> text "Qed."

instance CoqShow Proof where
  coqShow (ProofExact tm) = text "exact" <+> parens (coqShow tm) <> text "."
  coqShow (ProofString s) = text s
  coqShow _ = text "Other proof forms"


instance CoqShow Inductive where
  coqShow (Inductive bs) = text "Inductive"
    <+> hcat (punctuate (line <> text " with ") (map (nest 2 . coqShow ) bs)) <> dot

instance CoqShow InductiveBody where
  coqShow (InductiveBody name bs s cs) = hang 2 (coqShow name </> hsep (map coqShow bs) </> colon <+> coqShow s <+> text ":=") <$$> vsep (map coqShow cs)

instance CoqShow InductiveCtor where
  coqShow (InductiveCtor name bs mt) = text "| " <> text name <+> hsep (map coqShow bs) </> colon <+> coqShow mt

-- TODO: Is this how I should implement it...? If the two last thing have no sense, delete them.
instance CoqShow MatchItem where
  coqShow (MatchItem s mb mt) = coqShow s

instance CoqShow a => CoqShow (Maybe a) where
  coqShow Nothing  = text "_"
  coqShow (Just t) = coqShow t

-- TODO: Check whether this kind of pattern makes sense.
instance CoqShow Pattern where
  coqShow (PatCtor s ids) = coqShow s <+> hsep (map coqShow ids)
  coqShow (PatQualId s)   = coqShow s
  coqShow s               = text "Implement patterns."

instance CoqShow Sentence where
  coqShow (SentenceDefinition s) = coqShow s
  coqShow (SentenceInductive s)  = coqShow s
  coqShow (SentenceFixpoint s)   = coqShow s
  coqShow (SentenceLemma s)      = coqShow s
  coqShow _                      = text "TODO: Other sentence forms."

instance CoqShow Sort where
  coqShow Prop = text "Prop"
  coqShow Set  = text "Set"
  coqShow Type = text "Type"

-- TODO: All the same printing?
instance CoqShow SubstTy where
  coqShow xs = hsep (map (\x -> case x of
                                                          TermId s -> coqShow s
                                                          TermSubst s -> coqShow s
                                                          _ -> parens $ coqShow x) (sty_terms xs))

coqShowParens :: Term -> Doc
coqShowParens (TermId s)    = coqShow s
coqShowParens (TermSubst s) = coqShow s
coqShowParens s             = parens $ coqShow s

instance CoqShow Term where
  coqShow (TermImp s) = text "@" <> coqShow s
  coqShow (Todo s)           = text ("Todo." ++ s)
  coqShow (TermApp s ts)     = coqShowParens s <+> hsep (map (\x -> case x of
                                                          TermId s -> coqShow s
                                                          TermSubst s -> coqShow s
                                                          _ -> parens $ coqShow x) ts)
  coqShow (TermId s)         = coqShow s
  coqShow (TermFunction s t) = coqShow s <+> text "->" <+> coqShow t
  coqShow (TermSort s)       = coqShow s
  coqShow TermUnderscore   = text "_"
  coqShow (TermForall bs t)  = text "forall" <+> hsep (map coqShow bs) <> text "," <+> coqShow t
  coqShow (TermAbs bs t)     = text "fun" <+> hsep (map coqShow bs) <+> text "=>" <+> coqShow t
  coqShow (TermLet s t u)    = text "let" <+> coqShow t <+> text ":=" <+> coqShow s <+> text "in" <+> coqShow u
  coqShow (Tuple s)          = cTupled (map coqShow s)
  coqShow (TupleTy s)        = encloseSep empty empty (text " * ") (map (parens . coqShow) s)
  coqShow (TermMatch m _ s) = text "match" <+> coqShow m <+> text "with" <$$> vcat (map coqShow s) <$$> text "end" -- TODO: Version with Return clause
  coqShow (TermAnd s)  = encloseSep empty empty (text " /\\") (map (parens . coqShow) s)
  coqShow (TermEq s t) = coqShow s <+> text "=" <+> coqShow t
  coqShow (TermSubst s) = coqShow s
  coqShow s                  = text "TODO"
