{-# LANGUAGE FlexibleContexts      #-}
{-# LANGUAGE MultiParamTypeClasses #-}
module Autosubst.GeneratorLean (genLeanCode) where

import           Autosubst.Lean
import           Autosubst.GenM
import           Autosubst.Names
import           Autosubst.Syntax
import           Autosubst.TacticsLean
import           Autosubst.Types
import           Control.Monad.Except
import           Control.Monad.Reader
import           Control.Monad.State.Lazy
import           Data.List                as L

-- Generation of Lean syntax from a signature
genVar :: TId -> SubstTy -> [CBinder] -> GenM InductiveCtors
genVar x n b = do
  open_x <- isOpen x
  s <- finT_ x (sty_terms n)
  let t = [s] ==> sortType x n
  return [InductiveCtor (var_ x) b (Just t) | open_x]

genCons :: TId -> SubstTy -> [CBinder] -> Constructor -> GenM InductiveCtor
genCons x n b (Constructor cname pos) = do
  t <- do
    up_n_x <- mapM (\(Position bs y) -> liftM2 idApp (return y) (fmap sty_terms $ castUpSubst x (Position bs y) n)) pos
    return $  up_n_x ==> sortType x n
  return $ InductiveCtor (constructor_ cname) b (Just t)

genBody :: TId -> GenM InductiveBody
genBody x = do
  cs <- constructors x
  (n,b) <- introScopeVar "n" x
  varCons <- genVar x n b--(explicit_ b)
  constructors <- mapM (genCons x n b) cs--(explicit_ b)) cs
  return $ InductiveBody x [] (foldr TermFunction (TermSort Type) (getTypes (explicit_ b)) )  (varCons ++ constructors)

genBodies :: [TId] -> GenM Inductive
genBodies xs = fmap Inductive (mapM genBody xs)

-- Generation of congruences.
genCongruence :: TId -> Constructor -> GenM (Lemma, Lemma)
genCongruence x (Constructor cname pos) = do
    (m, bm) <- introScopeVar "m" x
    let s = getPattern "s" pos
    let t =  getPattern "t" pos
    let h =  getPattern "h" pos
    let bs s = mapM (\(y, Position binders arg) -> do
                                                  xs' <-  castUpSubst x (Position binders arg) m
                                                  return $ BinderImplicitNameType [y] (sortType arg xs')) (zip s pos)
    bes <- bs s
    bt <- bs t
    let eqs = zipWith TermEq (map TermId s) (map TermId t)
    let eqs' = zipWith BinderNameType (map (\s -> [s]) h) eqs
    let eq = TermEq (idApp cname (map TermId s)) (idApp cname (map TermId t))
    return (Lemma (congr_ cname) (bm ++ bes ++ bt) (eqs ==> eq) (ProofString "intros, congr; assumption"),
              Lemma (congr_ cname++"'") (bm ++ bes ++ bt ++ eqs') (eq) (ProofString "intros, congr; assumption"))

genCongruences :: TId -> GenM [(Lemma, Lemma)]
genCongruences x = join $ fmap (mapM (genCongruence x)) (constructors x)

genUpRenS :: TId -> TId -> GenM Definition
genUpRenS y z = do
  (mn, bs) <- introRenScopeS ("m", "n")
  (xi, bxi) <- genRenS "xi" mn
  return $ Definition (upRen_ y z) (bs ++ bxi) Nothing (upRenT y z xi)


traversal :: TId -> SubstTy -> (TId -> String) ->  (Term -> Term) -> Binders -> [SubstTy] -> (Term -> Term)  -> (TId -> [Term] -> Term) -> GenM FixpointLBody
traversal x scope name ret bargs args var_case sem = do
  let s = "s"
  --holes <- getHoles x
  cs <- constructors x
  open_x <- isOpen x
  let bindernms = concat (getNms bargs)
  let var_pattern = [Equation (PatCtorEx bindernms [PatCtor (var_ x) [s]]) (var_case (TermId s)) | open_x]
  let cons_pattern (Constructor cname pos) = do
                                            let holes = "_"
                                            let s = getPattern "s" pos
                                            args <- mapM (\pos -> mapM (castUpSubst x pos) args) pos
                                            return $ Equation (PatCtorEx bindernms [PatCtor cname (s)]) (sem cname
                                                        (zipWith (\(arg,s) (Position bs p) -> idApp (name p) (map TermSubst arg ++ [TermId s])) (zip args s) pos))
  cons_pat <- mapM cons_pattern cs
  let imp_b = bargs ++ [BinderNameType [s] (idApp x (sty_terms scope))]
  let ret_term = ret (TermId s)
  return $ FixpointLBody (name x) imp_b ret_term (var_pattern ++ cons_pat)


genRenaming :: TId -> GenM FixpointLBody
genRenaming x = do
  ((m,n),bs) <- introRenScope ("m", "n") x
  (xi,bxi) <- genRen x "xi" (m,n)
  toVarT <- toVar x xi
  traversal x m ren_ (const $ sortType x n) (bs ++ bxi) [xi]
                       (\s -> TermApp (var x n) [TermApp toVarT [s]])
                       (\x s -> idApp x ( s)) ---

genRenamings :: [TId] -> GenM FixpointL
genRenamings xs = fmap FixpointL (mapM genRenaming xs)

upSubstT :: TId -> TId -> SubstTy -> Term -> GenM Term
upSubstT y z m sigma = do
    pat <- patternSId z y
    m <- upSubst y [z] m
    let sigma' = sigma >>> idApp (ren_ z) pat
    return $ if y == z then TermApp cons_ [TermApp (var y m) [varZero_], sigma'] else sigma'

genUpS :: TId -> TId -> GenM Definition
genUpS y z = do
  ((m,n), bs) <- introSubstScopeS ("m", "n") z
  (sigma, b_sigma) <- genSubstS "sigma" (m,n) z
  sigma <- upSubstT y z n sigma
  return $ Definition (up_ y z) (bs ++ b_sigma) Nothing sigma

genSubstitution :: TId -> GenM FixpointLBody
genSubstitution x = do
  ((m, n), bmn) <- introRenScope ("m", "n") x
  (sigma, bs) <- genSubst x "sigma" (m,n)
  toVarT <- toVar x sigma
  traversal x m subst_ (const $ sortType x n) (bmn ++ bs) [sigma]
                       (\s -> TermApp toVarT [s])
                       (\x s -> idApp x s)

genSubstitutions :: [TId] -> GenM FixpointL
genSubstitutions xs = fmap FixpointL (mapM genSubstitution xs)

genUpId :: TId -> TId -> GenM Definition
genUpId y z = do
  (m, bm) <- introScopeVar "m" z
  m_var <- toVar z m
  (sigma, b_sigma) <- genSubstS "sigma" (m_var, m) z
  (eq, b_eq) <- genEq z "Eq" sigma (var z m)
  n <- tfresh "n"
  m <- upSubst z [y] m
  let ret = equiv_ (idApp (up_ y z) [sigma]) (var z m)
  shift <- patternSId z y
  let t' n = ap_ [idApp (ren_ z) shift, TermApp eq [n]]
  let t n = genPrfSimp (t' n)
  let u = if y == z then matchFin_ (TermId n) t eq_refl_ else t (TermId n)
  return $ Definition (upId_ y z) (implicit_ (bm ++ b_sigma) ++ b_eq) (Just ret) (TermAbs [BinderName n] u)

genIdL :: TId -> GenM FixpointLBody
genIdL x = do
  (m, bm) <- introScopeVar "m" x
  (sigma, bs) <- genSubst x "sigma" (m, m)
  xs <- substOf x
  eqs' <- mapM (\y -> liftM2 idApp (return $ var_ y) (fmap sty_terms (castSubst x y m))) xs
  (eqs, beqs) <- genEqs x "Eq" (sty_terms sigma) eqs' (\x y s -> return $ idApp (upId_ y x) [TermId "", s]) 
  toVarT <- toVar x eqs
  let ret s = TermEq (idApp (subst_ x) (sty_terms sigma ++ [s])) s
  traversal x m idSubst_ ret (bm ++ bs ++ beqs) [sigma, eqs]
                       (\s -> genPrfSimp (TermApp toVarT [s]))
                       (\x s -> genPrfSimp (idApp (congr_ x) s))

genIdSubst' :: TId -> GenM Lemma
genIdSubst' x = do
  (m, bm) <- introScopeVar "m" x
  (sigma, bs) <- genSubst x "sigma" (m, m)
  xs <- substOf x
  let s = "s"
  let s' = "s'"
  eqs' <- mapM (\y -> liftM2 idApp (return $ var_ y) (fmap sty_terms (castSubst x y m))) xs
  (eqs, beqs) <- genEqs x "Eq" (sty_terms sigma) eqs' (\x y s -> return $ idApp (upId_ y x) [TermUnderscore, s])
  let (eq_s, beq_s) = (TermId ("Eqs"++x), [BinderNameType ["Eqs"++x] (TermEq (TermId s) (TermId s'))])
  toVarT <- toVar x eqs
  let bds = implicit_ ((bm ++ bs ++ beqs) ++ [BinderNameType [s, s'] (idApp x (sty_terms m))])++ beqs ++ beq_s
  let ret = TermEq (idApp (subst_ x) (sty_terms sigma ++ [TermId s])) (TermId s')
  let prf = eqTrans_ (idApp (idSubst_ x) (map TermSubst [sigma, eqs]++[TermId s])) eq_s
  return $ Lemma (idSubst'_ x) bds ret (ProofExact prf)


genCompRenRen :: TId -> GenM FixpointLBody
genCompRenRen x = do
  ((k,l), bkl) <- introRenScope ("k", "l") x
  (m, bm) <- introScopeVar "m" x
  (xi, bxi) <- genRen x "xi" (m,k)
  (zeta,bzeta) <- genRen x "zeta" (k,l)
  (rho, brho) <- genRen x "rho" (m, l)
  (eqs, beqs) <- genEqs x "Eq" (zipWith (>>>) (sty_terms xi) (sty_terms zeta)) (sty_terms rho) (\x y s -> return$ if x == y then idApp up_ren_ren_ [TermId ("xi" ++ x), TermId ("zeta" ++ x), TermId ("rho" ++ x), s] else s)
  toVarT <- toVar x eqs
  let ret s = TermEq (idApp (ren_ x) (sty_terms zeta ++ [idApp (ren_ x) $ sty_terms xi  ++ [s]])) (idApp (ren_ x) (sty_terms rho ++ [s]))
  traversal x m compRenRen_ ret (bkl ++ bm ++ bxi ++ bzeta ++ brho ++ beqs) [xi, zeta, rho, eqs]
                  (\n -> genPrfSimp (ap_ [var x l, TermApp toVarT [n]]))
                  (\x s -> genPrfSimp (idApp (congr_ x) s))

genCompRenRen' :: TId -> GenM Lemma
genCompRenRen' x = do
  ((k,l), bkl) <- introRenScope ("k", "l") x
  (m, bm) <- introScopeVar "m" x
  (xi, bxi) <- genRen x "xi" (m,k)
  (zeta,bzeta) <- genRen x "zeta" (k,l)
  (rho, brho) <- genRen x "rho" (m, l)
  (eqs, beqs) <- genEqs x "Eq" (zipWith (>>>) (sty_terms xi) (sty_terms zeta)) (sty_terms rho) (\x y s -> return$ if x == y then idApp up_ren_ren_ [TermId ("xi" ++ x), TermId ("zeta" ++ x), TermId ("rho" ++ x), s] else s)
  toVarT <- toVar x eqs
  let s = "s"
  let s' = "s'"
  let (eq_s, beq_s) = (TermId ("Eqs"++x), [BinderNameType ["Eqs"++x] (TermEq (TermId s') (idApp (ren_ x) (sty_terms xi ++ [TermId s])))])
  let ret = TermEq (idApp (ren_ x) (sty_terms zeta ++ [TermId s'])) (idApp (ren_ x) (sty_terms rho ++ [TermId s]))
  let prf = ProofAssrt "h" TermUnderscore (idApp (compRenRen_ x)  (map TermSubst [xi, zeta, rho, eqs]++[TermId s]))
        [ProofString $ "rw Eqs"++x++", assumption"]
  let bds = implicit_ ((bkl ++ bm ++ bxi ++ bzeta ++ brho) ++ [BinderImplicitNameType [s] (idApp x (sty_terms m))]
        ++ [BinderImplicitNameType [s'] (idApp x (sty_terms k))])  ++ beqs ++ beq_s
  return $ Lemma (compRenRen'_ x) bds ret prf

genUpRenSubst :: TId -> TId -> GenM Definition
genUpRenSubst y z = do
  (k, bk) <- introScopeVarS "k"
  (l, bl) <- introScopeVarS "l"
  (m, bm) <- introScopeVar "m" z
  (xi, bxi) <- genRenS "xi" (k, l)
  (tau, btau) <- genSubstS "tau" (l, m) z
  (theta, btheta) <- genSubstS "theta" (k, m) z
  m_var <- toVar z m
  (eq, b_eq) <- genEq z "Eq" (xi >>> tau) theta
  n <- tfresh "n"
  m <- upSubst z [y] m
  let ret = equiv_ (idApp (upRen_ y z) [xi] >>> idApp (up_ y z) [tau] ) (idApp (up_ y z) [theta])
  shift <- patternSId z y
  let t n = ap_ [idApp (ren_ z) shift, TermApp eq [n]]
  let u = if y == z then matchFin_ (TermId n) t eq_refl_  else t (TermId n) -- TODO: Generalize the matchFIn thing.
  return $ Definition (up_ren_subst_ y z) (bk ++ bl ++ bm ++ bxi ++ btau ++ btheta ++ b_eq) (Just ret) (TermAbs [BinderName n] u)

genCompRenSubst :: TId -> GenM FixpointLBody
genCompRenSubst x = do
  ((k,l), bkl) <- introRenScope ("k", "l") x
  (m, bm) <- introScopeVar "m" x
  (xi, bxi) <- genRen x "xi" (m,k)
  (zeta,bzeta) <- genSubst x "tau" (k,l)
  (rho, brho) <- genSubst x "theta" (m, l)
  (eqs, beqs) <- genEqs x "Eq" (zipWith (>>>) (sty_terms xi) (sty_terms zeta)) (sty_terms rho) (\x y s -> return $ idApp (up_ren_subst_ y x) [TermId ("xi" ++ x), TermId ("tau" ++ x), TermId ("theta" ++ x), s])
  toVarT <- toVar x eqs
  let ret s = TermEq (idApp (subst_ x) (sty_terms zeta ++ [idApp (ren_ x) $ sty_terms xi  ++ [s]])) (idApp (subst_ x) (sty_terms rho ++ [s]))
  traversal x m compRenSubst_ ret (bkl ++ bm ++ bxi ++ bzeta ++ brho ++ beqs) [xi, zeta, rho, eqs]
                  (\n -> genPrfSimp (TermApp toVarT [n]))
                  (\x s -> genPrfSimp (idApp (congr_ x) s))

genCompRenSubst' :: TId -> GenM Lemma
genCompRenSubst' x = do
  ((k,l), bkl) <- introRenScope ("k", "l") x
  (m, bm) <- introScopeVar "m" x
  (xi, bxi) <- genRen x "xi" (m,k)
  (zeta,bzeta) <- genSubst x "tau" (k,l)
  (rho, brho) <- genSubst x "theta" (m, l)
  (eqs, beqs) <- genEqs x "Eq" (zipWith (>>>) (sty_terms xi) (sty_terms zeta)) (sty_terms rho) (\x y s -> return $ idApp (up_ren_subst_ y x) [TermId ("xi" ++ x), TermId ("tau" ++ x), TermId ("theta" ++ x), s])
  toVarT <- toVar x eqs
  let s = "s"
  let s' = "s'"
  let (eq_s, beq_s) = (TermId ("Eqs"++x), [BinderNameType ["Eqs"++x] (TermEq (TermId s') (idApp (ren_ x) (sty_terms xi ++ [TermId s])))])
  let ret = TermEq (idApp (subst_ x) (sty_terms zeta ++ [TermId s'])) (idApp (subst_ x) (sty_terms rho ++ [TermId s]))
  let prf = ProofAssrt "h" TermUnderscore (idApp (compRenSubst_ x)  (map TermSubst [xi, zeta, rho, eqs]++[TermId s]))
        [ProofString $ "rw Eqs"++x++", assumption"]
  let bds = (implicit_(bkl ++ bm ++ bxi ++ bzeta ++ brho) ++ [BinderImplicitNameType [s] (idApp x (sty_terms m))]
        ++ [BinderImplicitNameType [s'] (idApp x (sty_terms k))]) ++ beqs ++ beq_s
  return $ Lemma (compRenSubst'_ x) bds ret prf

genUpSubstRen :: TId -> TId -> GenM Definition
genUpSubstRen y z = do
  (k, bk) <- introScopeVarS "k"
  (l, bl) <- introScopeVar "l" z
  (m, bm) <- introScopeVar "m" z
  (sigma, bsigma) <- genSubstS "sigma" (k, l) z
  (zeta, bzeta) <- genRen z "zeta" (l, m)
  (theta, btheta) <- genSubstS "theta" (k, m) z
  (eq, b_eq) <- genEq z "Eq" (sigma >>> idApp (ren_ z) (sty_terms zeta)) theta
  n <- tfresh "n"
  m <- upSubst z [y] m
  zeta' <- upSubst z [y] zeta
  pat <- patternSId z y
  let ret = equiv_ (idApp (up_ y z) [sigma] >>> idApp (ren_ z) (sty_terms zeta')) (idApp (up_ y z) [theta])
  shift <- patternSId z y
  let t n = eqTrans_ (idApp (compRenRen_ z) (pat ++ sty_terms zeta' ++ zipWith (>>>) (sty_terms zeta) pat ++ map (const (TermAbs [BinderName "x"] eq_refl_)) pat ++ [ TermApp sigma [n]]))
                (eqTrans_ (eqSym_ (idApp (compRenRen_ z) (sty_terms zeta ++ pat ++ zipWith (>>>) (sty_terms zeta) pat ++ map (const ((TermAbs [BinderName "x"] eq_refl_))) pat ++ [ TermApp sigma [n]])))
                (ap_ [idApp (ren_ z) pat, TermApp eq [n]]))
  let u = if y == z then matchFin_ (TermId n) t eq_refl_ else t (TermId n)
  return $ Definition (up_subst_ren_ y z) (bk ++ bl ++ bm ++ bsigma ++ bzeta ++ btheta ++ b_eq) (Just ret) (TermAbs [BinderName n] u)

genCompSubstRen :: TId -> GenM FixpointLBody
genCompSubstRen x = do
  ((k,l), bkl) <- introRenScope ("k", "l") x
  (m, bm) <- introScopeVar "m" x
  (sigma, bsigma) <- genSubst x "sigma" (m,k)
  (zeta,bzeta) <- genRen x "zeta" (k,l)
  (theta, btheta) <- genSubst x "theta" (m, l)
  xs <- substOf x
  sigmazeta <- mapM (\(y, sigma) -> do
                zeta' <- castSubst x y zeta
                return $ sigma >>> idApp (ren_ y) (sty_terms zeta')) (zip xs  (sty_terms sigma))
  (eqs, beqs) <- genEqs x "Eq" sigmazeta (sty_terms theta) (\z y s -> do
                                  sigma' <- toVar z sigma
                                  zeta' <- castSubst x z zeta
                                  theta' <- toVar z  theta
                                  return $ idApp (up_subst_ren_ y z) ([sigma'] ++ sty_terms zeta' ++[theta', s]))
  toVarT <- toVar x eqs
  let ret s = TermEq (idApp (ren_ x) (sty_terms zeta ++ [idApp (subst_ x) $ sty_terms sigma  ++ [s]])) (idApp (subst_ x) (sty_terms theta ++ [s]))
  traversal x m compSubstRen_ ret (bkl ++ bm ++ bsigma ++ bzeta ++ btheta ++ beqs) [sigma, zeta, theta, eqs]
                  (\n -> genPrfSimp (TermApp toVarT [n]))
                  (\x s -> genPrfSimp (idApp (congr_ x) s))

genCompSubstRen' :: TId -> GenM Lemma
genCompSubstRen' x = do
  ((k,l), bkl) <- introRenScope ("k", "l") x
  (m, bm) <- introScopeVar "m" x
  (sigma, bsigma) <- genSubst x "sigma" (m,k)
  (zeta,bzeta) <- genRen x "zeta" (k,l)
  (theta, btheta) <- genSubst x "theta" (m, l)
  xs <- substOf x
  sigmazeta <- mapM (\(y, sigma) -> do
                zeta' <- castSubst x y zeta
                return $ sigma >>> idApp (ren_ y) (sty_terms zeta')) (zip xs  (sty_terms sigma))
  (eqs, beqs) <- genEqs x "Eq" sigmazeta (sty_terms theta) (\z y s -> do
                                  sigma' <- toVar z sigma
                                  zeta' <- castSubst x z zeta
                                  theta' <- toVar z  theta
                                  return $ idApp (up_subst_ren_ y z) ([sigma'] ++ sty_terms zeta' ++[theta', s]))
  let s = "s"
  let s' = "s'"
  let (eq_s, beq_s) = (TermId ("Eqs"++x), [BinderNameType ["Eqs"++x] (TermEq (TermId s') (idApp (subst_ x) (sty_terms sigma ++ [TermId s])))])
  let ret = TermEq (idApp (ren_ x) (sty_terms zeta ++ [TermId s'])) (idApp (subst_ x) (sty_terms theta ++ [TermId s]))
  let prf = ProofAssrt "h" TermUnderscore (idApp (compSubstRen_ x)  (map TermSubst [sigma, zeta, theta, eqs]++[TermId s]))
        [ProofString $ "rw Eqs"++x++", assumption"]
  let bds = (implicit_ (bkl ++ bm ++ bsigma ++ bzeta ++ btheta) ++ [BinderImplicitNameType [s] (idApp x (sty_terms m))]
        ++ [BinderImplicitNameType [s'] (idApp x (sty_terms k))]) ++ beqs ++ beq_s
  return $ Lemma (compSubstRen'_ x) bds ret prf

genUpSubstSubst :: TId -> TId -> GenM Definition
genUpSubstSubst y z = do
  (k, bk) <- introScopeVarS "k"
  (l, bl) <- introScopeVar "l" z
  (m, bm) <- introScopeVar "m" z
  (sigma, bsigma) <- genSubstS "sigma" (k, l) z
  (tau, btau) <- genSubst z "tau" (l, m)
  (theta, btheta) <- genSubstS "theta" (k, m) z
  (eq, b_eq) <- genEq z "Eq" (sigma >>> idApp (subst_ z) (sty_terms tau)) theta
  n <- tfresh "n"
  m <- upSubst z [y] m
  zeta' <- upSubst z [y] tau
  pat <- patternSId z y
  let ret = equiv_ (idApp (up_ y z) [sigma] >>> idApp (subst_ z) (sty_terms zeta')) (idApp (up_ y z) [theta])
  shift <- patternSId z y
  zs <- substOf z
  pat' <- mapM (\(tau, z') -> do
                      p' <- castSubst z z' (SubstScope pat)
                      return $ tau >>> (idApp (ren_ z') (sty_terms p')) ) (zip (sty_terms tau) zs)
  let t n = (eqTrans_ (idApp (compRenSubst_ z) (pat ++ sty_terms zeta'  ++ (map (\(z, p) -> p >>> z) (zip (sty_terms zeta') pat)) ++ map (const ((TermAbs [BinderName "x"] eq_refl_))) pat ++ [ TermApp sigma [n]]))
                (eqTrans_ (eqSym_ (idApp (compSubstRen_ z) (sty_terms tau ++ pat ++ pat' ++ map (const ((TermAbs [BinderName "x"] eq_refl_))) pat ++ [ TermApp sigma [n]])))
                (ap_ [(idApp (ren_ z) pat), TermApp eq [n]])))
  let u = if y == z then matchFin_ (TermId n) t eq_refl_ else t (TermId n)
  return $ Definition (up_subst_subst_ y z) (bk ++ bl ++ bm ++ bsigma ++ btau ++ btheta ++ b_eq) (Just ret) (TermAbs [BinderName n] u)
--TODO: implicit args


genCompSubstSubst :: TId -> GenM FixpointLBody
genCompSubstSubst x = do
  ((k,l), bkl) <- introRenScope ("k", "l") x
  (m, bm) <- introScopeVar "m" x
  (sigma, bsigma) <- genSubst x "sigma" (m,k)
  (tau,btau) <- genSubst x "tau" (k,l)
  (theta, btheta) <- genSubst x "theta" (m, l)
  xs <- substOf x
  sigmatau <- mapM (\(y, sigma) -> do
                tau' <- castSubst x y tau
                return $ sigma >>> idApp (subst_ y) (sty_terms tau')) (zip xs  (sty_terms sigma))
  (eqs, beqs) <- genEqs x "Eq" sigmatau (sty_terms theta) (\z y s -> do
                                  sigma' <- toVar z sigma
                                  tau' <- castSubst x z tau
                                  theta' <- toVar z  theta
                                  return $ idApp (up_subst_subst_ y z) ([sigma'] ++ sty_terms tau' ++[theta', s]))
  toVarT <- toVar x eqs
  let ret s = TermEq (idApp (subst_ x) (sty_terms tau ++ [idApp (subst_ x) $ sty_terms sigma  ++ [s]])) (idApp (subst_ x) (sty_terms theta ++ [s]))
  traversal x m compSubstSubst_ ret (bkl ++ bm ++ bsigma ++ btau ++ btheta ++ beqs) [sigma, tau, theta, eqs]
                  (\n -> genPrfSimp (TermApp toVarT [n]))
                  (\x s -> genPrfSimp (idApp (congr_ x) s))

genCompSubstSubst' :: TId -> GenM Lemma
genCompSubstSubst' x = do
  ((k,l), bkl) <- introRenScope ("k", "l") x
  (m, bm) <- introScopeVar "m" x
  (sigma, bsigma) <- genSubst x "sigma" (m,k)
  (tau,btau) <- genSubst x "tau" (k,l)
  (theta, btheta) <- genSubst x "theta" (m, l)
  xs <- substOf x
  sigmatau <- mapM (\(y, sigma) -> do
                tau' <- castSubst x y tau
                return $ sigma >>> idApp (subst_ y) (sty_terms tau')) (zip xs  (sty_terms sigma))
  (eqs, beqs) <- genEqs x "Eq" sigmatau (sty_terms theta) (\z y s -> do
                                  sigma' <- toVar z sigma
                                  tau' <- castSubst x z tau
                                  theta' <- toVar z  theta
                                  return $ idApp (up_subst_subst_ y z) ([sigma'] ++ sty_terms tau' ++[theta', s]))
  let s = "s"
  let s' = "s'"
  let (eq_s, beq_s) = (TermId ("Eqs"++x), [BinderNameType ["Eqs"++x] (TermEq (TermId s') (idApp (subst_ x) (sty_terms sigma ++ [TermId s])))])
  let ret = TermEq (idApp (subst_ x) (sty_terms tau ++ [TermId s'])) (idApp (subst_ x) (sty_terms theta ++ [TermId s]))
  let prf = ProofAssrt "h" TermUnderscore (idApp (compSubstSubst_ x)  (map TermSubst [sigma, tau, theta, eqs]++[TermId s]))
        [ProofString $ "rw Eqs"++x++", assumption"]
  let bds = (implicit_ (bkl ++ bm ++ bsigma ++ btau ++ btheta) ++ [BinderImplicitNameType [s] (idApp x (sty_terms m))]
        ++ [BinderImplicitNameType [s'] (idApp x (sty_terms k))]) ++ beqs ++ beq_s
  return $ Lemma (compSubstSubst'_ x) bds ret prf

genUpExtRen :: TId -> TId -> GenM Definition
genUpExtRen y z = do
  (m, bm) <- introScopeVarS "m"
  (n, bn) <- introScopeVarS "n"
  (xi, bxi) <- genRenS "xi" (m, n)
  (zeta,bzeta) <- genRenS "zeta" (m,n)
  (eq, b_eq) <- genEq z "Eq" xi zeta
  let ret = equiv_ (idApp (upRen_ y z) [xi]) (idApp (upRen_ y z) [zeta])
  n <- tfresh "n"
  let t n = TermApp eq [n]
  let s = matchFin_ (TermId n) (\n -> ap_ [shift_, t n]) eq_refl_
  let u = if y == z then s else t (TermId n)
  return $ Definition (upExtRen_ y z) (bm ++ bn ++ bxi ++ bzeta ++ b_eq) (Just ret) (TermAbs [BinderName n] u)

genExtRen :: TId -> GenM FixpointLBody
genExtRen x = do
  ((m,n), bmn) <- introRenScope ("m", "n") x
  (xi, bxi) <- genRen x "xi" (m,n)
  (zeta, bzeta) <- genRen x "zeta" (m,n)
  xs <- substOf x
  (eqs, beqs) <- genEqs x "Eq" (sty_terms xi) (sty_terms zeta) (\x y s -> return$  idApp (upExtRen_ y x) [TermId ("xi" ++ x), TermId ("zeta" ++ x), s]) -- TODO: Shouldn't this want SubsttTy instead of Terms?
  let ret s = TermEq (idApp (ren_ x) (sty_terms xi ++ [s])) (idApp (ren_ x) (sty_terms zeta ++ [s]))
  toVarT <- toVar x eqs
  traversal x m extRen_ ret (bmn ++ bxi ++ bzeta ++ beqs) [xi, zeta, eqs]
                  (\z -> genPrfSimp (ap_ [var x n, TermApp toVarT [z]]))
                  (\x s -> genPrfSimp (idApp (congr_ x) s))

genExtRen' :: TId -> GenM Lemma
genExtRen' x = do
  ((m,n), bmn) <- introRenScope ("m", "n") x
  (xi, bxi) <- genRen x "xi" (m,n)
  (zeta, bzeta) <- genRen x "zeta" (m,n)
  xs <- substOf x
  let s = "s"
  let s' = "s'"
  (eqs, beqs) <- genEqs x "Eq" (sty_terms xi) (sty_terms zeta) (\x y s -> return$  idApp (upExtRen_ y x) [TermId ("xi" ++ x), TermId ("zeta" ++ x), s])
  let (eq_s, beq_s) = (TermId ("Eqs"++x), [BinderNameType ["Eqs"++x] (TermEq (TermId s) (TermId s'))])
  let ret = TermEq (idApp (ren_ x) (sty_terms xi ++ [TermId s])) (idApp (ren_ x) (sty_terms zeta ++ [TermId s']))
  let bds = (implicit_ (bmn ++ bxi ++ bzeta) ++ [BinderImplicitNameType [s, s'] (idApp x (sty_terms m))]) ++ beqs ++ beq_s
  let prftm = idApp (extRen_ x) (map TermSubst [xi, zeta, eqs]++[TermId s])
  let prf = ProofAssrt "h" TermUnderscore prftm [ProofString $ "rw <- Eqs"++x++", assumption"]
  return $ Lemma (extRen'_ x) bds ret prf


genUpExt :: TId -> TId -> GenM Definition
genUpExt y z = do
  (m, bm) <- introScopeVarS "m"
  (n, bn) <- introScopeVar "n" z
  (sigma, bsigma) <- genSubstS "sigma" (m, n) z
  (tau,btau) <- genSubstS "tau" (m,n) z
  (eq, b_eq) <- genEq z "Eq" sigma tau
  let ret = equiv_ (idApp (up_ y z) [sigma]) (idApp (up_ y z) [tau])
  shift <- patternSId z y
  n <- tfresh "n"
  let t n = ap_ [idApp (ren_ z) shift, TermApp eq [n]]
  let u = if y == z then matchFin_ (TermId n) t eq_refl_ else t (TermId n)
  return $ Definition (upExt_ y z) (bm ++ bn ++ bsigma ++ btau ++ b_eq) (Just ret) (TermAbs [BinderName n] u)

genExt :: TId -> GenM FixpointLBody
genExt x = do
  ((m,n), bmn) <- introRenScope ("m", "n") x
  (sigma, bsigma) <- genSubst x "sigma" (m,n)
  (tau, btau) <- genSubst x "tau" (m,n)
  xs <- substOf x
  (eqs, beqs) <- genEqs x "Eq" (sty_terms sigma) (sty_terms tau) (\x y s -> return$  idApp (upExt_ y x) [TermId ("sigma" ++ x), TermId ("tau" ++ x), s]) -- TODO: Shouldn't this want SubsttTy instead of Terms?
  let ret s = TermEq (idApp (subst_ x) (sty_terms sigma ++ [s])) (idApp (subst_ x) (sty_terms tau ++ [s]))
  toVarT <- toVar x eqs
  traversal x m ext_ ret (bmn ++ bsigma ++ btau ++ beqs) [sigma, tau, eqs]
                  (\z -> genPrfSimp (TermApp toVarT [z]))   -- TODO: I didn't need the renaming case for Up. Dive into that.
                  (\x s -> genPrfSimp (idApp (congr_ x) s))

genExt' :: TId -> GenM Lemma
genExt' x = do
  ((m,n), bmn) <- introRenScope ("m", "n") x
  (sigma, bsigma) <- genSubst x "sigma" (m,n)
  (tau, btau) <- genSubst x "tau" (m,n)
  xs <- substOf x
  let s = "s"
  let s' = "s'"
  (eqs, beqs) <- genEqs x "Eq" (sty_terms sigma) (sty_terms tau) (\x y s -> return$  idApp (upExt_ y x) [TermId ("sigma" ++ x), TermId ("tau" ++ x), s])
  let (eq_s, beq_s) = (TermId ("Eqs"++x), [BinderNameType ["Eqs"++x] (TermEq (TermId s) (TermId s'))])
  let ret = TermEq (idApp (subst_ x) (sty_terms sigma ++ [TermId s])) (idApp (subst_ x) (sty_terms tau ++ [TermId s']))
  let bds = (implicit_ (bmn ++ bsigma ++ btau) ++ [BinderImplicitNameType [s, s'] (idApp x (sty_terms m))]) ++ beqs ++ beq_s
  let prftm = idApp (ext_ x) (map TermSubst [sigma, tau, eqs]++[TermId s])
  let prf = ProofAssrt "h" TermUnderscore prftm [ProofString $ "rw <- Eqs"++x++", assumption"]
  return $ Lemma (ext'_ x) bds ret prf


genUpRinstInst :: TId -> TId -> GenM Definition
genUpRinstInst y z = do
  (m, bm) <- introScopeVarS "m"
  (n, bn) <- introScopeVar "n" z
  n_var <- toVar z n
  (xi, bxi) <- genRenS "xi" (m, n_var)
  (sigma, bsigma) <- genSubstS "sigma" (m, n) z
  (eq, b_eq) <- genEq z "Eq" (xi >>> var z n) sigma
  n' <- upSubst z [y] n
  let ret = equiv_ (idApp (upRen_ y z) [xi] >>> var z n') (idApp (up_ y z) [sigma])
  shift <- patternSId z y
  let t n = ap_ [idApp (ren_ z) shift, TermApp eq [n]]

  -- insert simplification step
  h <- tfresh "h"
  let s t n = TermProof $ ProofAssrt h TermUnderscore (t n) [ProofString $ "simp with substLemmas at *, assumption"]
  n <- tfresh "n"
  let u = if y == z then matchFin_ (TermId n) (s t) eq_refl_ else (s t) (TermId n)
  return $ Definition (up_rinstInst_ y z) (bm ++ bn ++ bxi ++ bsigma ++ b_eq) (Just ret) (TermAbs [BinderName n] u)
--TODO upRinstInst'

genRinstInst :: TId -> GenM FixpointLBody
genRinstInst x = do
  ((m,n), bmn) <- introRenScope ("m", "n") x
  (xi, bxi) <- genRen x "xi" (m,n)
  (sigma, bsigma) <- genSubst x "sigma" (m,n)
  xs <- substOf x
  xis' <- mapM (\(y, xi) -> do
    n' <- castSubst x y n
    return$  xi >>> var y n') (zip xs (sty_terms xi))
  (eqs, beqs) <- genEqs x "Eq" xis' (sty_terms sigma) (\x y s -> return $  idApp (up_rinstInst_ y x) [TermId ("xi" ++ x), TermId ("sigma" ++ x), s]) -- TODO: Make this last part general: Allow a function whihc takes SubstTy arguments and then automatically gives right terms.
  let ret s = TermEq (idApp (ren_ x) (sty_terms xi ++ [s])) (idApp (subst_ x) (sty_terms sigma ++ [s]))
  toVarT <- toVar x eqs
  traversal x m rinstInst_ ret (bmn ++ bxi ++ bsigma ++ beqs) [xi, sigma, eqs]
                  (\n -> genPrfSimp (TermApp toVarT [n]))
                  (\x s -> genPrfSimp (idApp (congr_ x ) s))

genRinstInst' :: TId -> GenM Lemma
genRinstInst' x = do
  ((m,n), bmn) <- introRenScope ("m", "n") x
  (xi, bxi) <- genRen x "xi" (m,n)
  (sigma, bsigma) <- genSubst x "sigma" (m,n)
  xs <- substOf x
  xis' <- mapM (\(y, xi) -> do
    n' <- castSubst x y n
    return$  xi >>> var y n') (zip xs (sty_terms xi))

  let s = "s"
  let s' = "s'"
  (eqs, beqs) <- genEqs x "Eq" xis' (sty_terms sigma) (\x y s -> return $  idApp (up_rinstInst_ y x) [TermId ("xi" ++ x), TermId ("sigma" ++ x), s]) -- TODO: Make this last part general: Allow a function whihc takes SubstTy arguments and then automatically gives right terms.

  let (eq_s, beq_s) = (TermId ("Eqs"++x), [BinderNameType ["Eqs"++x] (TermEq (TermId s) (TermId s'))])
  let ret = TermEq (idApp (ren_ x) (sty_terms xi ++ [TermId s'])) (idApp (subst_ x) (sty_terms sigma ++ [TermId s]))
  let bds = (implicit_ (bmn ++ bsigma ++ bxi) ++ [BinderImplicitNameType [s, s'] (idApp x (sty_terms m))]) ++ beqs ++ beq_s
  let prftm = idApp (rinstInst_ x) (map TermSubst [xi, sigma, eqs]++[TermId s])
  let prf = ProofAssrt "h" TermUnderscore prftm [ProofString $"rw <- Eqs"++x++", assumption"]
  return $ Lemma (rinstInst'_ x) bds ret prf

fext_ :: String
fext_ = "funext"

--
genRinstInstL :: TId -> GenM Lemma
genRinstInstL x = do
  ((m,n), bmn) <- introRenScope ("m", "n") x
  (xi, bxi) <- genRen x "xi" (m,n)
  xs <- substOf x
  let ret = TermEq (idApp (ren_ x) (sty_terms xi )) (idApp (subst_ x) [idApp "comp" $ [var x n]++sty_terms xi])
  let proof = idApp fext_ [TermAbs [BinderName "x"] (idApp (rinstInst_ x) ((sty_terms xi )++ [TermUnderscore] ++(map (const (TermAbs [BinderName "n"] eq_refl_)) xs)++[TermId "x"] ))]
  return $ Lemma (rinstInstL_ x) (bmn++bxi) ret (ProofExact proof)


genLemmaInstId :: TId -> GenM Lemma
genLemmaInstId x = do
  (m, bm) <- introScopeVar "m" x
  xs <- substOf x
  vars <- mapM (\y -> liftM2 idApp (return $ var_ y) (liftM sty_terms (castSubst x y m))) xs
  let ret = TermEq (idApp (subst_ x) vars) id_
  let proof = idApp fext_ [TermAbs [BinderName "x"] (idApp (idSubst_ x) ( vars ++ (map (const (TermAbs [BinderName "n"] eq_refl_)) xs) ++ [TermApp id_ [TermId "x"]]))]
  return $ Lemma (instId_ x) bm ret (ProofExact proof)

genLemmaRinstId :: TId -> GenM Lemma
genLemmaRinstId x = do
    (m, bm) <- introScopeVar "m" x

    xs <- substOf x
    vars <- mapM (\y -> liftM2 idApp (return $ var_ y) (liftM sty_terms (castSubst x y m))) xs
    let ret = TermEq (idApp ("@"++(ren_ x)) [TermSubst m, TermSubst m, id_]) id_
    let proof = ProofExact (eqTrans_ (TermApp (TermId $ rinstInstL_ x)  [id_])  (TermId $ instId_ x) )
    return $ Lemma (rinstId_ x) bm ret (ProofExact $ TermProof proof)


genLemmaVarL :: TId -> GenM Lemma
genLemmaVarL x = do
  ((m,n), bmn) <- introRenScope ("m", "n") x
  (sigma,bsigma) <- genSubst x "sigma" (m,n)
  sigma' <- toVar x sigma
  let ret = TermEq ((var x m) >>> (idApp (subst_ x) (sty_terms sigma))) sigma'
  let proof = idApp fext_ [TermAbs [BinderName "x"] eq_refl_]
  return $ Lemma (varL_ x) (bmn ++ bsigma) ret (ProofExact proof)

genLemmaVarLRen :: TId -> GenM Lemma
genLemmaVarLRen x = do
  ((m,n), bmn) <- introRenScope ("m", "n") x
  (xi,bxi) <- genRen x "xi" (m,n)
  sigma' <- toVar x xi
  let ret = TermEq ((var x m) >>> (idApp (ren_ x) (sty_terms xi))) ( idApp "comp" $ [var x n]++sty_terms xi )
  let proof = idApp fext_ [TermAbs [BinderName "x"] eq_refl_]
  return $ Lemma (varLRen_ x) (bmn ++ bxi) ret (ProofExact proof)



genLemmaSubstComp :: TId -> GenM (Lemma, Lemma)
genLemmaSubstComp x = do
  ((k,l), bkl) <- introRenScope ("k", "l") x
  (m, bm) <- introScopeVar "m" x
  (sigma, bsigma) <- genSubst x "sigma" (m,k)
  (tau,btau) <- genSubst x "tau" (k,l)
  xs <- substOf x
  let s = "s"
  sigmatau <- mapM (\(y, sigma) -> do
                tau' <- castSubst x y tau
                return $ sigma >>> idApp (subst_ y) (sty_terms tau')) (zip xs  (sty_terms sigma))
  let ret = TermEq (idApp (subst_ x) (sty_terms tau ++ [idApp (subst_ x) $ sty_terms sigma  ++ [TermId s]])) (idApp (subst_ x) (sigmatau ++ [TermId s]))
  let proof = idApp (compSubstSubst_ x) (sty_terms sigma ++ sty_terms tau ++ map (const TermUnderscore) xs ++ map (const (TermAbs [BinderName "n"] eq_refl_)) xs ++ [TermId s])
  let ret' = TermEq ((idApp (subst_ x) (sty_terms sigma)) >>> (idApp (subst_ x) (sty_terms tau))) (idApp (subst_ x) sigmatau) -- (sty_terms sigma >>> idApp (subst_ x) (sty_terms tau)))
  let proof' = idApp fext_ [TermAbs [BinderName "n"] (idApp (compComp_ x) (sty_terms sigma ++ sty_terms tau ++ [TermId "n"]))]
  return (Lemma (compComp_ x) (bkl ++ bm ++ bsigma ++ btau ++ [BinderNameType [s] (idApp x (sty_terms m))]) ret (ProofExact proof),
          Lemma (compComp'_ x) (bkl ++ bm ++ bsigma ++ btau) ret' (ProofExact proof'))

genLemmaSRComp :: TId -> GenM (Lemma, Lemma) --(Lemma, [Lemma])
genLemmaSRComp x = do
    ((k,l), bkl) <- introRenScope ("k", "l") x
    (m, bm) <- introScopeVar "m" x
    (sigma, bsigma) <- genSubst x "sigma" (m,k)
    (tau,btau) <- genRen x "xi" (k,l)
    xs <- substOf x
    let s = "s"
    sigmatau <- mapM (\(y, sigma) -> do
                          tau' <- castSubst x y tau
                          return $ sigma >>> idApp (ren_ y) (sty_terms tau')) (zip xs  (sty_terms sigma))
    let ret = TermEq (idApp (ren_ x) (sty_terms tau ++ [idApp (subst_ x) $ sty_terms sigma  ++ [TermId s]])) (idApp (subst_ x) (sigmatau ++ [TermId s]))
    let proof = idApp (compSubstRen_ x) (sty_terms sigma ++ sty_terms tau ++ map (const TermUnderscore) xs ++ map (const (TermAbs [BinderName "n"] eq_refl_)) xs ++ [TermId s])
    let ret' = TermEq ((idApp (subst_ x) (sty_terms sigma)) >>> (idApp (ren_ x) (sty_terms tau))) (idApp (subst_ x) sigmatau) -- (sty_terms sigma >>> idApp (subst_ x) (sty_terms tau)))
    let proof' = idApp fext_ [TermAbs [BinderName "n"] (idApp (compSR_ x) (sty_terms sigma ++ sty_terms tau ++ [TermId "n"]))]
    return (Lemma (compSR_ x) (bkl ++ bm ++ bsigma ++ btau ++ [BinderNameType [s] (idApp x (sty_terms m))]) ret (ProofExact proof),
            Lemma (compSR'_ x) (bkl ++ bm ++ bsigma ++ btau) ret' (ProofExact proof'))

genLemmaRSComp :: TId -> GenM (Lemma, Lemma) --(Lemma, [Lemma])
genLemmaRSComp x = do
    ((k,l), bkl) <- introRenScope ("k", "l") x
    (m, bm) <- introScopeVar "m" x
    (sigma, bsigma) <- genRen x "xi" (m,k)
    (tau,btau) <- genSubst x "sigma" (k,l)
    xs <- substOf x
    let s = "s"
    sigmatau <- mapM (\(y, sigma) -> do
              tau' <- castSubst x y tau
              return $ idApp "comp" $ (sty_terms tau')++(sty_terms tau)) (zip xs  (sty_terms sigma))
    let ret = TermEq (idApp (subst_ x) (sty_terms tau ++ [idApp (ren_ x) $ sty_terms sigma  ++ [TermId s]])) (idApp (subst_ x) ([idApp "comp" $ (sty_terms tau)++(sty_terms sigma)] ++ [TermId s]))
    let proof = idApp (compRenSubst_ x) (sty_terms sigma ++ sty_terms tau ++ map (const TermUnderscore) xs ++ map (const (TermAbs [BinderName "n"] eq_refl_)) xs ++ [TermId s])
    let ret' = TermEq ((idApp (ren_ x) (sty_terms sigma)) >>> (idApp (subst_ x) (sty_terms tau))) (idApp (subst_ x) [idApp "comp" $ (sty_terms tau)++(sty_terms sigma)]) -- (sty_terms sigma >>> idApp (subst_ x) (sty_terms tau)))
    let proof' = idApp fext_ [TermAbs [BinderName "n"] (idApp (compRS_ x) (sty_terms sigma ++ sty_terms tau ++ [TermId "n"]))]
    return (Lemma (compRS_ x) (bkl ++ bm ++ bsigma ++ btau ++ [BinderNameType [s] (idApp x (sty_terms m))]) ret (ProofExact proof),
            Lemma (compRS'_ x) (bkl ++ bm ++ bsigma ++ btau) ret' (ProofExact proof'))

genLemmaRRComp :: TId -> GenM (Lemma, Lemma) --(Lemma, [Lemma])
genLemmaRRComp x = do
    ((k,l), bkl) <- introRenScope ("k", "l") x
    (m, bm) <- introScopeVar "m" x
    (sigma, bsigma) <- genRen x "xi" (m,k)
    (tau,btau) <- genRen x "rho" (k,l)
    xs <- substOf x
    let s = "s"
    sigmatau <- mapM (\(y, sigma) -> do
                          tau' <- castSubst x y tau
                          return $ idApp "comp" $ (sty_terms tau')++(sty_terms tau)) (zip xs  (sty_terms sigma))
    let ret = TermEq (idApp (ren_ x) (sty_terms tau ++ [idApp (ren_ x) $ sty_terms sigma  ++ [TermId s]])) (idApp (ren_ x) ([idApp "comp" $ (sty_terms tau)++(sty_terms sigma)] ++ [TermId s]))
    let proof = idApp (compRenRen_ x) (sty_terms sigma ++ sty_terms tau ++ map (const TermUnderscore) xs ++ map (const (TermAbs [BinderName "n"] eq_refl_)) xs ++ [TermId s])
    let ret' = TermEq ((idApp (ren_ x) (sty_terms sigma)) >>> (idApp (ren_ x) (sty_terms tau))) (idApp (ren_ x) [idApp "comp" $ (sty_terms tau)++(sty_terms sigma)]) -- (sty_terms sigma >>> idApp (subst_ x) (sty_terms tau)))
    let proof' = idApp fext_ [TermAbs [BinderName "n"] (idApp (compRR_ x) (sty_terms sigma ++ sty_terms tau ++ [TermId "n"]))]
    return (Lemma (compRR_ x) (bkl ++ bm ++ bsigma ++ btau ++ [BinderNameType [s] (idApp x (sty_terms m))]) ret (ProofExact proof),
            Lemma (compRR'_ x) (bkl ++ bm ++ bsigma ++ btau) ret' (ProofExact proof'))
-- meta definitions: is and destruct

isDeConstr :: TId -> Constructor -> GenM [MetaDefinition]
isDeConstr x (Constructor cname pos) = do
    let cargs = map ExprId (getPattern "s" pos)
    h <- (getHoles x)
    let h' = concatMap (const " _") h
    -- constructor
    let is1 = genIsEq' (cname++h', cargs)
    let de1 = genDeEq' x (cname++h', cargs)
    let is2 = genIsEq' ("congr_"++cname++"'", cargs)
    let de2 = genDeEq' x ("congr_"++cname++"'", cargs)
    -- lemmas
    idL <- genIdSubst' x
    let is3 = genRecIsEq x cname idL
    let de3 = genRecDeEq x cname idL

    let tactic_type pos = TermApp (TermId "tactic")
                [TupleTy $ foldr (\s t -> t++[TermId "expr"]) [] pos]
    let retTuple = ExprTuple $ map (const (ExprId "e")) (L.findIndices (const True) cargs)

    return [MetaDefinition (is_ cname x) [] (TermFunction expr bool)
      ([is1,is2,is3]++defIs),
      MetaDefinition (destruct_ cname x) [] (TermFunction expr (tactic_type pos))
        ([de1,de2,de3]++[ExprEq  (ExprId "e") (ExprRet retTuple)])]

isDeConstrs :: TId -> GenM [MetaDefinition]
isDeConstrs x = do
  l <- join $ fmap (mapM (isDeConstr x)) (constructors x)
  let l' = concat l
  return l'

isDeInstRen :: TId -> GenM [MetaDefinition]
isDeInstRen x = do
  let nm = "InstRen"
  l2 <- genCompRenRen' x
--  l1 <- genExtRen' x
  let isEqs = [genRecIsEq x nm l2]
  let deEqs = [genRecDeEq x nm l2]
  let defis = [ExprEq (ExprId "e") (eIdApp (is_ "ren" x) [ExprId "e"])]
  let defde = [ExprEq (ExprId "e") (eIdApp (destruct_ "ren" x)[ExprId "e"])]

  return [MetaDefinition (is_ nm x) [] (TermFunction expr bool) (isEqs++defis),
          MetaDefinition (destruct_ nm x) []
          (TermFunction expr $ TermApp (TermId "tactic") [TupleTy [expr, expr]]) (deEqs++defde)]

isDeRen :: TId -> GenM [MetaDefinition]
isDeRen x = do
  let nm = "ren"
  let args = [ExprId "xi", ExprId "s"]
  l1 <- genExtRen' x
  let isEq2 = genIsEq l1
  let isEq1 = ExprEq (ExprQuote $ ExprApp (ExprId "ren_tm") (map ExprQp args)) tt
  let deEq2 = genDeEq nm l1
  let deEq1 = ExprEq (ExprQuote $ ExprApp (ExprId "ren_tm") (map ExprQp args))
                                  (ExprRet $ ExprTuple args)
  return [MetaDefinition (is_ nm x) [] (TermFunction expr bool) ([isEq1, isEq2]++defIs),
    MetaDefinition (destruct_ nm x) []
    (TermFunction expr $ TermApp (TermId "tactic") [TupleTy [expr, expr]]) ([deEq1, deEq2]++defDe)]

isDeSubst :: TId -> GenM [MetaDefinition]
isDeSubst x = do
  let nm = "subst"
  let args = [ExprId "xi", ExprId "s"]
  l1 <- genExt' x
  let isEq1 = ExprEq (ExprQuote $ ExprApp (ExprId $ nm++sep++x) (map ExprQp args)) tt
  let isEq2 = ExprEq (ExprQuote $ ExprApp (ExprId $ "ren"++sep++x) (map ExprQp args)) tt
  let isEq3 = genIsEq l1
  let deEq1 = ExprEq (ExprQuote $ ExprApp (ExprId $ nm++sep++x) (map ExprQp args))
                                    (ExprRet $ ExprTuple args)
  let deEq2 = ExprEq (ExprQuote $ ExprApp (ExprId $ "ren"++sep++x) (map ExprQp args))
                                    (ExprRet $ ExprTuple args)
  let deEq3 = genDeEq nm l1
  return [MetaDefinition (is_ nm x) [] (TermFunction expr bool) ([isEq1,isEq2,deEq3]++defIs),
    MetaDefinition (destruct_ nm x) []
    (TermFunction expr $ TermApp (TermId "tactic") [TupleTy [expr, expr]]) ([deEq1,deEq2,deEq3]++defDe)]





isDeInstSubst :: TId -> GenM [MetaDefinition] --TODO: treating renamings also as subst now
isDeInstSubst x = do
  let nm = "Inst"
  l0 <- genExt' x
  l1 <- genCompRenRen' x
  l2 <- genCompRenSubst' x
  l3 <- genCompSubstRen' x
  l4 <- genCompSubstSubst' x
  l5 <- genIdSubst' x
  l6 <- genRinstInst' x
  let isEqs = map (genRecIsEq x nm) [l1,l2,l3,l4,l5,l6]
  let deEqs =[genDeEq nm l0]++(map (genRecDeEq x nm) [l1,l2,l3,l4,l5,l6])
  let defis = [ExprEq (ExprId "e") (eIdApp (is_ "subst" x) [ExprId "e"])] --TODO
  let defde = [ExprEq (ExprId "e") (eIdApp (destruct_ "subst" x)[ExprId "e"])] --TODO
  return [MetaDefinition (is_ nm x) [] (TermFunction expr bool) (isEqs++defis),
          MetaDefinition (destruct_ nm x) []
          (TermFunction expr $ idApp "tactic" [TupleTy [expr, expr]])(deEqs++defde)]

isDeConsCompId :: [Definition] -> [TId] -> GenM [MetaDefinition]
isDeConsCompId d xs = do
  let ps =  [ExprId "H1", ExprId "H2"]
  let nms = ["scons", "comp"]++map (\x -> var_ x) xs
  let isEqs name = map ((\(s,ps)-> ExprEq (ExprQuote $ ExprApp (ExprId s)
                (map ExprQp ps))(ExprApp (ExprId (is_ name "")) [ExprId $ "Eq"])) . argsD)
              d++[ExprEq (ExprQuote $ ExprApp (ExprId $ name) (map ExprQp ps))tt]++defIs

  let deEqs name args = map ((\(s,ps)-> ExprEq (ExprQuote $ ExprApp (ExprId s)
                (map ExprQp ps))(ExprApp (ExprId (destruct_ name "")) [ExprId "Eq"])) . argsD) d++
                  [ExprEq (ExprQuote $ ExprApp (ExprId $ name) (map ExprQp args))(ExprRet $ ExprTuple args)]

  let is = map (\nm -> MetaDefinition (is_ nm "") [] (TermFunction expr bool) (isEqs nm)) nms
  let deScons = MetaDefinition  (destruct_ "scons" "") [] (TermFunction expr $ TermApp (TermId "tactic") [TupleTy [expr, expr]])
                                (deEqs "scons" ps ++defDe)
  let deComp = MetaDefinition (destruct_ "comp" "") [] (TermFunction expr $ TermApp (TermId "tactic") [TupleTy [expr, expr]])
                                                                (deEqs "comp" ps++defDe)
  let deVar = map (\x -> MetaDefinition (destruct_ (var_ x) "") [] (TermFunction expr $ TermApp (TermId "tactic") [expr])
                                  (deEqs (var_ x) [ExprId "s"]++[ExprEq (ExprQuote $ ExprQp $ ExprId "e") (ExprRet $ ExprId "e")])) xs
  return $ is++ [deScons, deComp] ++ deVar

normInstRen :: TId -> GenM Expr
normInstRen x = do
  let e n = ExprId ("e"++n)
  let qe n = ExprQp (e n)
  let t1 = eIdApp (is_ (var_ x) "") [e "1"]
  let caseIdL = ExprIte t1 $ ExprDQuote (eIdApp (idSubst'_ x) [qe "1",qe "2"])

  let t2 = eIdApp (is_ "lam" x)[e "2"] --TODO generalize
  let caseLam = ExprIte t2 (ExprDo
        [ExprLet (e"3") (eIdApp (destruct_ "lam" x) [e"2"]),
        ExprLet (e"4") (eIdApp (nmUp_ x) [e"1", e"1"]), ExprLet (e"5") (eIdApp (nmRinst_ x) [e"3", e"2"]),
        ExprDQuote (eIdApp ("congr_lam'")  [qe "4",qe "5"])])
  let t3 f = eIdApp (f (ren_ x) "") [e"1"]
  let caseInst = ExprIte (t3 is_) $ ExprDo
        [ExprLet (ExprId "(e3,e4)") (t3 destruct_), ExprLet (e"5") (eIdApp (nmCmp_ x) [e"4",e"1"] ),
        ExprLet (e"6") (eIdApp (nmRinst_ x) [e"3",e"5"] ),
        ExprDQuote (eIdApp (compRenRen'_ x)  [qe"5", qe "2",qe "6"])]

  let caseRefl = ExprDQuote $ eIdApp (extRen'_ x) (map ExprQp [e "1", e "2"])

  return $ caseIdL $ caseLam $ caseInst $  caseRefl

normInstSubst :: TId -> GenM Expr
normInstSubst x = do
  let e n = ExprId ("e"++n)
  let qe n = ExprQp (e n)
  let t f nm n = eIdApp (f nm "") [e n]
  let caseIdL = ExprIte (t is_ (var_ x) "1") $ ExprDQuote (eIdApp (idSubst'_ x) [qe "1",qe "2"])
--TODO instId benutzen?

  let caseLam = ExprIte (t is_ ("lam"++sep++x) "2") (ExprDo
        [ExprLet (e"3") (eIdApp (destruct_ "lam" x) [e"2"]),
        ExprLet (e"4") (eIdApp (nmUp_ x) [e"1", e"1"]), ExprLet (e"5") (eIdApp (nmInst_ x) [e"3", e"2"]),
        ExprDQuote (eIdApp ("congr_lam'")  [qe "4",qe "5"])])

  let caseInst = ExprIte (t is_ (subst_ x) "2") $ ExprDo
        [ExprLet (ExprId "(e3,e4)") (t destruct_ (subst_ x) "2"), ExprLet (e"5") (eIdApp (nmCmp_ x) [e"4",e"1"] ),
        ExprLet (e"6") (eIdApp (nmInst_ x) [e"3",e"5"] ),
        ExprDQuote (eIdApp (compSubstSubst'_ x)  [qe"5", qe "2",qe "6"])]


--  let caseCompComp = ExprIte (t is_ (subst_ x) "2") $ ExprDo
      -- TODO : lemma verstärken, richtung?

  let caseRefl = ExprDQuote $ eIdApp (ext'_ x) [qe "1",qe "2"]
  return $ caseIdL $ caseLam $ caseInst $ caseRefl


normComp :: TId -> GenM Expr
normComp x = do
    let e n = ExprId ("e"++n)
    let qe n = ExprQp (e n)
    let t f nm n = eIdApp (f nm "")[e n]
    let caseCons = ExprIte (t is_ "scons" "1") $ ExprDo
          [ExprLet (ExprId "(e3,e4)") (t destruct_ "scons" "1"), ExprLet (e"5") (eIdApp (nmInst_ x) [e"3",e"2"] ),
          ExprLet (e"6") (eIdApp (nmCmp_ x) [e"4",e"2"] ), ExprLet (e"7") (eIdApp (nmCons_ x) [e"5",e"6"] ),
          ExprDQuote (eIdApp ("scons_comp'")  [qe"5", qe "1",qe "6"])]
    let caseVarL = ExprIte (ExprBAnd (t is_ ("subst_"++x) "1") (t is_ ("var_"++x) "2")) $ ExprDo
          [ExprLet (ExprId "(e3,e4)") (t destruct_ ("subst_"++x) "1"),
          ExprDQuote (eIdApp ("varL_"++ x)  [qe"3"])]
    let caseRefl = ExprDQuote $ eIdApp ("comp'") [qe"1",qe"2"]
    return $ caseCons $ caseVarL $ caseRefl

normUp :: TId -> GenM Expr
normUp x = do
  return $ ExprId ""

genNormalizeSubterm :: TId -> GenM MetaDefinition
genNormalizeSubterm x = do
  rinst <- normInstRen x
  inst <- normInstSubst x
  cmp <- normComp x
  up <- normUp x
  let h nm = ExprBEq (ExprId "n") (ExprId nm)

  let ite = ExprIte (h "Inst") inst $ ExprIte (h "Rinst") rinst $
          ExprIte (h "Comp") cmp -- $ ExprIte (h "Up") up
        (ExprDQuote $ ExprId "%%e1=%%e2") --TODO

  return $ MetaDefinition ("normalize"++sep++x) [] ((==>) [nat, expr, expr] (TermId "tactic expr"))
            [ExprEq (ExprId "n e1 e2") ite]

-- MISSING: All the inferred lemmas.
genCodeT :: [TId] -> [(TId,TId)] -> GenM [Sentence]
genCodeT xs upList = do
  varSorts <- filterM isOpen xs -- Sorts which have variables
  substSorts <- filterM (const $ return True) xs -- FEATURE: Only sorts which require renaming / substitution.
  types <- genBodies xs
  congruences' <- mapM genCongruences xs
  let congruences = concatMap (\(c1, c2) -> [c1] ++ [c2]) (map unzip congruences')
  upRen <- mapM (uncurry genUpRenS) upList
  renamings <- genRenamings substSorts
  ups <- mapM (uncurry genUpS) upList
  substitutions <- genSubstitutions substSorts
  upId <- mapM (uncurry genUpId) upList
  idlemmas <- mapM genIdL substSorts
  extUpRen <- mapM (uncurry genUpExtRen) upList
  extRen <- mapM genExtRen substSorts
  extUp <- mapM (uncurry genUpExt) upList
  ext <- mapM genExt substSorts
  compRenRen <- mapM genCompRenRen substSorts
  upRenSubst <- mapM (uncurry genUpRenSubst) upList
  compRenSubst <- mapM genCompRenSubst substSorts
  upSubstRen <- mapM (uncurry genUpSubstRen) upList
  compSubstRen <- mapM genCompSubstRen substSorts
  upSubstSubst <- mapM (uncurry genUpSubstSubst) upList
  compSubstSubst <- mapM genCompSubstSubst substSorts
  upRinstInst <- mapM (uncurry genUpRinstInst) upList
  rinstInst <- mapM genRinstInst substSorts
  lemmaInstId <- mapM genLemmaInstId substSorts
  lemmaRinstId <- mapM genLemmaRinstId substSorts
  rinstInstL <- mapM genRinstInstL substSorts
  lemmaVarL <- mapM genLemmaVarL varSorts
  lemmaVarLRen <- mapM genLemmaVarLRen varSorts
  lemmaSubstComp <- mapM genLemmaSubstComp substSorts
  lemmaSRComp <- mapM genLemmaSRComp substSorts
  lemmaRSComp <- mapM genLemmaRSComp substSorts
  lemmaRRComp <- mapM genLemmaRRComp substSorts
  let (l1, l2) = unzip lemmaSubstComp
  let (l3, l4) = unzip lemmaSRComp
  let (l5, l6) = unzip lemmaRSComp
  let (l7, l8) = unzip lemmaRRComp
  --transitivity lemmas
  idlemmas' <- mapM genIdSubst' substSorts
  extRen' <- mapM genExtRen' substSorts
  ext' <- mapM genExt' substSorts
  compRenRen' <- mapM genCompRenRen' substSorts
  compRenSubst' <- mapM genCompRenSubst' substSorts
  compSubstRen' <- mapM genCompSubstRen' substSorts
  compSubstSubst' <- mapM genCompSubstSubst' substSorts
  rinstInst' <- mapM genRinstInst' substSorts
  --meta defs
  --lemmas:
  isDeConstr <- mapM (\x-> isDeConstrs x ) xs
  isDeInstSubst <- mapM (\x-> isDeInstSubst x) substSorts
  isDeInstRen <- mapM (\x-> isDeInstRen x) substSorts
  isDeRen <- mapM (\x-> isDeRen x) substSorts
  isDeSubst <- mapM (\x-> isDeSubst x) substSorts
  let upLems =  concat [upRenSubst, upSubstRen, upSubstSubst]
  isDeConsComp <- isDeConsCompId upLems substSorts
  normalizeSubterm <- mapM genNormalizeSubterm xs


  return $ [SentenceInductive types] ++ map SentenceLemma (concat congruences) -- FEATURE: Different intermediate values. --}
          ++ map SentenceDefinition upRen ++ [SentenceFixpointL renamings]
          ++ map SentenceDefinition ups ++ [SentenceFixpointL substitutions]
          ++ map SentenceDefinition upId ++ [SentenceFixpointL $ FixpointL idlemmas]
          ++ map SentenceLemma idlemmas'
          ++ map SentenceDefinition extUpRen ++ [SentenceFixpointL $ FixpointL extRen] ++ map SentenceLemma extRen'
          ++ map SentenceDefinition extUp ++ [SentenceFixpointL $ FixpointL ext] ++ map SentenceLemma ext'
          ++ [SentenceFixpointL $ FixpointL compRenRen] ++ map SentenceLemma compRenRen'
          ++ map SentenceDefinition upRenSubst ++ [SentenceFixpointL $ FixpointL compRenSubst]
          ++ map SentenceLemma compRenSubst'
          ++ map SentenceDefinition upSubstRen ++ [SentenceFixpointL $ FixpointL compSubstRen]
          ++ map SentenceLemma compSubstRen'
          ++ map SentenceDefinition upSubstSubst ++ [SentenceFixpointL $ FixpointL compSubstSubst]
          ++ map SentenceLemma compSubstSubst'
          ++ map SentenceDefinition upRinstInst ++ [SentenceFixpointL $ FixpointL rinstInst]
          ++  map SentenceLemma rinstInst'
          ++ map SentenceLemma (rinstInstL++lemmaInstId ++ lemmaRinstId ++ lemmaVarL ++ lemmaVarLRen ++ l1 ++ l2 ++ l3 ++ l4 ++ l5 ++ l6 ++ l7 ++ l8)
          --TODO lemmas'
          ++ map SentenceMetaDefinition (concat isDeConstr)
          ++ map SentenceMetaDefinition (concat isDeRen)++ map SentenceMetaDefinition (concat isDeSubst)
          ++ map SentenceMetaDefinition (concat isDeInstRen)++ map SentenceMetaDefinition (concat isDeInstSubst)
          ++ map SentenceMetaDefinition isDeConsComp
          ++ map SentenceMetaDefinition normalizeSubterm

genLeanCode :: [[TId]] -> GenM Sentences
genLeanCode xs = do
  let all = [(x,y) | x <- concat xs, y <- concat xs]
  (_, code) <- foldM (\(ups,sentences) x -> do
                                  xs <- substOf (L.head x)
                                  let up_x = [(x, y) | x <- xs, y <- xs]
                                  code_x <- genCodeT x (L.filter (`elem` ups) up_x)
                                  return (L.filter (`notElem` up_x) ups, sentences ++ [code_x]))
                                  (all, []) xs
  return $ concat code
